/*===========================================================================

 Copyright (C) 2018-2019 Tetsuo Koyama

 This file is a part of GetFEM++

 GetFEM++  is  free software;  you  can  redistribute  it  and/or modify it
 under  the  terms  of the  GNU  Lesser General Public License as published
 by  the  Free Software Foundation;  either version 3 of the License,  or
 (at your option) any later version along with the GCC Runtime Library
 Exception either version 3.1 or (at your option) any later version.
 This program  is  distributed  in  the  hope  that it will be useful,  but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 or  FITNESS  FOR  A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 License and GCC Runtime Library Exception for more details.
 You  should  have received a copy of the GNU Lesser General Public License
 along  with  this program;  if not, write to the Free Software Foundation,
 Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301, USA.

===========================================================================*/

/**@file elastodynamic_side_boundary.cc
   @brief 2.5D plan wave elastodynamic analysis problem.
*/
#include <math.h>
#include "getfem/getfem_mesher.h"
#include "getfem/getfem_regular_meshes.h"
#include "getfem/getfem_generic_assembly.h"
#include "getfem/getfem_assembling.h"
#include "getfem/getfem_models.h"
#include "getfem/getfem_model_solvers.h"
#include "gmm/gmm.h"
#include "gmm/gmm_inoutput.h"

using std::cout;
using std::endl;
using bgeot::scalar_type; /* = double */
using bgeot::dim_type;
using bgeot::size_type;   /* = unsigned long */
using bgeot::base_small_vector; /* special class for small (dim<16) vectors */
using bgeot::base_node; /* geometrical nodes () */

/* definition of some matrix/vector types. These ones are built
 * using the predefined types in Gmm++
 */
typedef gmm::rsvector<scalar_type> sparse_vector_type;
typedef gmm::row_matrix<sparse_vector_type> sparse_matrix_type;
typedef gmm::csc_matrix<double> csc_matrix_type;
typedef std::vector<scalar_type> plain_vector;

scalar_type sol_t;
/* right hand side */
scalar_type sol_f(const base_node &x){
  scalar_type sol_f = 0.0;
  return sol_f;
}

/*
  structure for the 2.5D plan wave elastodynamic analysis problem
*/
struct elastodynamic_side_boundary_problem {
  scalar_type LX, LY; /* geometrical parameter */
  int NX, NY, NZ;

  enum { ALL = -1, LEFT = 1, RIGHT = 2, TOP = 3, BOTTOM = 4 };

  scalar_type clambda; /* First Lame coefficient (N/m^2) */
  scalar_type cmu;     /* Second Lame coefficient (N/m^2) */
  scalar_type nu;      /* Poisson's ratio */
  scalar_type rho;     /* Mass Density*/
  scalar_type Vp;      /* P wave velocity */
  scalar_type Vs;      /* S wave velocity */
  scalar_type Cvh;
  scalar_type Dx;

  scalar_type sol_t;
  scalar_type thetav;
  scalar_type thetah;
  scalar_type thetap;
  scalar_type thetas;

  getfem::mesh mesh1;     /* 1.5D mesh */
  getfem::mesh mesh;     /* 2.5D mesh */
  getfem::mesh_im mim;   /* integration methods. */
  getfem::mesh_fem m2f_ui, m2f_uo; /* the vector mesh_fem, for the 2.5D Plane Wave solution */
  getfem::mesh_fem m1f_ui, m1f_uo; /* the vector mesh_fem, for the 1.5D Plane Wave solution */
  getfem::mesh_fem m2f_d; /* the scalor mesh_fem, for the 2.5D Plane Wave solution */
  getfem::mesh_fem m1f_d; /* the scalor mesh_fem, for the 1.5D Plane Wave solution */
  getfem::mesh_fem m2f_rhsi; /* the mesh_fem for the right hand side(f(x),..) */
  getfem::mesh_fem m2f_rhso;
  getfem::mesh_fem m1f_rhsi;
  getfem::mesh_fem m1f_rhso;

  scalar_type elements_degree /* elemens degree */;

  scalar_type dt, T, beta, gamma; /* parameter of Newmark scheme */
  scalar_type alpha;
  scalar_type residual;        /* max residual for the iterative solvers     */

  plain_vector external_force_for_Dot_u; /* External Force Coefficient for Dot_u */

  getfem::ga_workspace workspace;

  std::string datafilename;
  bgeot::md_param PARAM;

  int init(void);
  bool solve(void);
  elastodynamic_side_boundary_problem(void) :
  mim(mesh),
  m2f_ui(mesh, 2), m2f_uo(mesh, 1), m1f_ui(mesh, 1), m1f_uo(mesh, 1), m2f_d(mesh, 1), m1f_d(mesh, 1),
  m2f_rhsi(mesh, 2), m2f_rhso(mesh, 1), m1f_rhsi(mesh, 1), m1f_rhso(mesh, 1) {}

};

/* Read parameters from the .param file, build the mesh, set finite element
 * and integration methods and selects the boundaries.
 */
int elastodynamic_side_boundary_problem::init(void) {

  std::string FEM_TYPE = PARAM.string_value("FEM_TYPE", "FEM name");
  std::string INTEGRATION = PARAM.string_value("INTEGRATION",
                                               "Name of integration method");

  /* Physical parameters */
  cmu = PARAM.real_value("cmu", "Second Lame coefficient (N/m^2)");
  nu  = PARAM.real_value("nu", "Poisson's ratio");
  rho = PARAM.real_value("rho", "Mass Density");
  clambda = 2.0*cmu*nu/(1.0-2.0*nu);
  Vp = sqrt((clambda+2.0*cmu)/rho);
  Vs = sqrt(cmu/rho);

  /* Geometrical parameters */
  thetav = PARAM.real_value("thetav", "Input Wave Tilt");
  thetah = PARAM.real_value("thetah", "Input Wave Azimuth");
  if(thetav == 0.0 || thetah == 90.0) {
    Cvh = 0.0;
  } else {
    Cvh = Vs/sin(thetav)/cos(thetah);
  }
  Dx = -sin(thetav)*cos(thetah)/Vs;
  thetas = thetav;
  thetap = asin(Vp/Vs*sin(thetas));

  /* output filename */
  datafilename = PARAM.string_value("ROOTFILENAME","Base name of data files.");

  /* First step : build the mesh */
  cout << "Mesh generation\n";
  LX = PARAM.real_value("LX", "Size in X");
  LY = PARAM.real_value("LY", "Size in Y");
  NX = PARAM.int_value("NX", "number of space steps ");
  NY = PARAM.int_value("NY", "number of space steps ");
  NZ = PARAM.int_value("NZ", "number of space steps ");
  std::vector<base_node> fixed;
  double hx = LX/NX;
  double hy = LY/NY;
  base_node org(0.0, 0.0);
  std::vector<bgeot::base_small_vector> vect(2);
  vect[0] = bgeot::base_small_vector(hx, 0.0);
  vect[1] = bgeot::base_small_vector(0.0, hy);
  std::vector<int> ref(2);
  ref[0] = NX;
  ref[1] = NY;
  getfem::parallelepiped_regular_mesh(
    mesh, 2, org, vect.begin(), ref.begin()
  );
  getfem::pfem pf_u = getfem::fem_descriptor(FEM_TYPE);
  getfem::pintegration_method ppi = getfem::int_method_descriptor(INTEGRATION);

  /* set the region of mesh */
  cout << "Selecting regions in boundaries\n";
  getfem::mesh_region border_faces;
  getfem::outer_faces_of_mesh(mesh, border_faces);

  getfem::mesh_region fbox
    = getfem::select_faces_in_box(mesh, border_faces, base_node(0.01, 0.01),
                                  base_node(LX-0.01, LY-0.01));

  getfem::mesh_region fright
    = getfem::select_faces_of_normal(mesh, border_faces,
                                     base_small_vector(+1.0, 0.0), 0.01);

  getfem::mesh_region fleft
    = getfem::select_faces_of_normal(mesh, border_faces,
                                     base_small_vector(-1.0, 0.0), 0.01);

  getfem::mesh_region ftop
    = getfem::select_faces_of_normal(mesh, border_faces,
                                     base_small_vector( 0.0,+1.0), 0.01);

  getfem::mesh_region fbottom
    = getfem::select_faces_of_normal(mesh, border_faces,
                                     base_small_vector( 0.0,-1.0), 0.01);

  mesh.region( RIGHT) = getfem::mesh_region::subtract( fright, fbox);
  mesh.region(  LEFT) = getfem::mesh_region::subtract(  fleft, fbox);
  mesh.region(   TOP) = getfem::mesh_region::subtract(   ftop, fbox);
  mesh.region(BOTTOM) = getfem::mesh_region::subtract(fbottom, fbox);

  mesh.write_to_file(datafilename + ".msh2");
  getfem::vtk_export mesh_exp(datafilename + ".vtk", false);

  /* set the finite element on the m2f_u and m2f_d */
  m1f_ui.set_finite_element(mesh.convex_index(), pf_u);
  m1f_uo.set_finite_element(mesh.convex_index(), pf_u);
  m1f_d.set_finite_element(mesh.convex_index(), pf_u);

  m2f_ui.set_finite_element(mesh.convex_index(), pf_u);
  m2f_uo.set_finite_element(mesh.convex_index(), pf_u);
  m2f_d.set_finite_element(mesh.convex_index(), pf_u);

  m2f_rhsi.set_finite_element(mesh.convex_index(), pf_u);
  m2f_rhso.set_finite_element(mesh.convex_index(), pf_u);
  m1f_rhsi.set_finite_element(mesh.convex_index(), pf_u);
  m1f_rhso.set_finite_element(mesh.convex_index(), pf_u);

  mim.set_integration_method(mesh.convex_index(), ppi);

  /* set the solve parameter */
  beta = PARAM.real_value("BETA", "Newmark theme beta parameter");
  gamma = PARAM.real_value("GAMMA", "Newmark theme gamma parameter");
  dt = PARAM.real_value("DT", "Time step");

  residual = PARAM.real_value("RESIDUAL");
  if (residual == 0.) residual = 1e-10;
  T = PARAM.real_value("T", "final time");

  return 0;
};

bool elastodynamic_side_boundary_problem::solve(void) {

  getfem::model model;

  // Material Parameter
  model.add_initialized_scalar_data("lambda", clambda);
  model.add_initialized_scalar_data("mu", cmu);
  model.add_initialized_scalar_data("rho", rho);
  model.add_initialized_scalar_data("Vp", Vp);
  model.add_initialized_scalar_data("Vs", Vs);

  // Geometoric parameter
  model.add_initialized_scalar_data("Cvh", Cvh);
  model.add_initialized_scalar_data("Dx", Dx);

  // Main unkown of the problem
  getfem::partial_mesh_fem m1f_ui1(m1f_ui);
  getfem::partial_mesh_fem m1f_uo1(m1f_uo);
  getfem::partial_mesh_fem m1f_ui2(m1f_ui);
  getfem::partial_mesh_fem m1f_uo2(m1f_uo);
  dal::bit_vector kept_dofi1 = m1f_ui1.basic_dof_on_region(LEFT);
  dal::bit_vector kept_dofo1 = m1f_uo1.basic_dof_on_region(LEFT);
  dal::bit_vector kept_dofi2 = m1f_ui2.basic_dof_on_region(RIGHT);
  dal::bit_vector kept_dofo2 = m1f_uo2.basic_dof_on_region(RIGHT);
  m1f_ui1.adapt(kept_dofi1);
  m1f_uo1.adapt(kept_dofo1);
  m1f_ui2.adapt(kept_dofi2);
  m1f_uo2.adapt(kept_dofo2);
  model.add_fem_variable("ux1", m1f_ui1);
  model.add_fem_variable("uz1", m1f_ui1);
  model.add_fem_variable("uo1", m1f_uo1);
  model.add_fem_variable("ux2", m1f_ui2);
  model.add_fem_variable("uz2", m1f_ui2);
  model.add_fem_variable("uo2", m1f_uo2);
  model.add_fem_variable("ui", m2f_ui);
  model.add_fem_variable("uo", m2f_uo);

  // Newmark beta theme
  getfem::add_Newmark_scheme(model, "ux1", beta, gamma);
  getfem::add_Newmark_scheme(model, "uz1", beta, gamma);
  getfem::add_Newmark_scheme(model, "uo1", beta, gamma);
  getfem::add_Newmark_scheme(model, "ux2", beta, gamma);
  getfem::add_Newmark_scheme(model, "uz2", beta, gamma);
  getfem::add_Newmark_scheme(model, "uo2", beta, gamma);
  getfem::add_Newmark_scheme(model, "ui", beta, gamma);
  getfem::add_Newmark_scheme(model, "uo", beta, gamma);

  // Stiffness brick

  getfem::add_generic_elliptic_brick(model, mim, "ux1", "mu");
  getfem::add_generic_elliptic_brick(model, mim, "ux2", "mu");
  getfem::add_generic_elliptic_brick(model, mim, "uz1", "mu");
  getfem::add_generic_elliptic_brick(model, mim, "uz2", "mu");
  getfem::add_isotropic_linearized_elasticity_brick(model, mim, "ui", "lambda", "mu");
  getfem::add_generic_elliptic_brick(model, mim, "uo1", "mu");
  getfem::add_generic_elliptic_brick(model, mim, "uo2", "mu");
  getfem::add_generic_elliptic_brick(model, mim, "uo", "mu");

  // Damping matrix on Dot_ui and Dot_uo.

  // TODO Assembling and add term
  getfem::add_nonlinear_term(model, mim, "lambda*Grad_Dot_ux1*Test_Dot_uz1", LEFT, false);

  // Mass brick.

  getfem::add_mass_brick(model, mim, "Dot2_ux1", "rho-Dx*Dx*(lambda+2.0*mu)");
  getfem::add_mass_brick(model, mim, "Dot2_ux2", "rho-Dx*Dx*(lambda+2.0*mu)");
  getfem::add_mass_brick(model, mim, "Dot2_uz1", "rho-Dx*Dx*mu");
  getfem::add_mass_brick(model, mim, "Dot2_uz2", "rho-Dx*Dx*mu");
  getfem::add_mass_brick(model, mim, "Dot2_uo1", "rho-Dx*Dx*(lambda+2.0*mu)");
  getfem::add_mass_brick(model, mim, "Dot2_uo2", "rho-Dx*Dx*(lambda+2.0*mu)");
  getfem::add_mass_brick(model, mim, "Dot2_ui", "rho-Dx*Dx*(lambda+2.0*mu)");
  getfem::add_mass_brick(model, mim, "Dot2_uo", "rho-Dx*Dx*(lambda+2.0*mu)");

  // Damper at TOP

  getfem::add_linear_term(model, mim, "(rho*Vs*Dot_ux1).Test_Dot_ux1", TOP);
  getfem::add_linear_term(model, mim, "(rho*Vp*Dot_uz1).Test_Dot_uz1", TOP);
  getfem::add_linear_term(model, mim, "(rho*Vs*Dot_ux2).Test_Dot_ux2", TOP);
  getfem::add_linear_term(model, mim, "(rho*Vp*Dot_uz2).Test_Dot_uz2", TOP);
  model.add_fixed_size_data("rhoV3", 2);
  model.set_real_variable("rhoV3")[0] = rho*Vs;
  model.set_real_variable("rhoV3")[1] = rho*Vp;
  getfem::add_linear_term(model, mim, "(rhoV3.*Dot_ui).Dot_ui", TOP);
  getfem::add_linear_term(model, mim, "(rho*Vs*Dot_uo1).Test_Dot_uo1", TOP);
  getfem::add_linear_term(model, mim, "(rho*Vs*Dot_uo2).Test_Dot_uo2", TOP);
  getfem::add_linear_term(model, mim, "(rho*Vs*Dot_uo).Test_Dot_uo", TOP);

  // Damper at LEFT
  model.add_fixed_size_data("rhoV1", 2);
  model.set_real_variable("rhoV1")[0] = rho*Vp;
  model.set_real_variable("rhoV1")[1] = rho*Vs;
  getfem::add_linear_term(model, mim, "(rhoV1.*Dot_ui).Dot_ui", LEFT);
  getfem::add_linear_term(model, mim, "(rho*Dot_uo).Test_Dot_uo");

  // Damper at RIGHT
  model.add_fixed_size_data("rhoV2", 2);
  model.set_real_variable("rhoV2")[0] = rho*Vp;
  model.set_real_variable("rhoV2")[1] = rho*Vs;
  getfem::add_linear_term(model, mim, "(rhoV2.*Dot_ui).Dot_ui", RIGHT);
  getfem::add_linear_term(model, mim, "(rho*Vs*Dot_uo).Test_Dot_uo");

  // Null initial value.
  plain_vector U;

  gmm::resize(U, m1f_ui1.nb_dof()); gmm::clear(U); gmm::copy(U, model.set_real_variable("ux1"));
  gmm::resize(U, m1f_ui1.nb_dof()); gmm::clear(U); gmm::copy(U, model.set_real_variable("uz1"));
  gmm::resize(U, m1f_ui2.nb_dof()); gmm::clear(U); gmm::copy(U, model.set_real_variable("ux2"));
  gmm::resize(U, m1f_ui2.nb_dof()); gmm::clear(U); gmm::copy(U, model.set_real_variable("uz2"));
  gmm::resize(U, m1f_uo1.nb_dof()); gmm::clear(U); gmm::copy(U, model.set_real_variable("uo1"));
  gmm::resize(U, m1f_uo2.nb_dof()); gmm::clear(U); gmm::copy(U, model.set_real_variable("uo2"));

  gmm::resize(U, m2f_ui.nb_dof()); gmm::clear(U); gmm::copy(U, model.set_real_variable("ui"));
  gmm::resize(U, m2f_uo.nb_dof()); gmm::clear(U); gmm::copy(U, model.set_real_variable("uo"));

  plain_vector V;

  gmm::resize(V, m1f_ui1.nb_dof()); gmm::clear(V); gmm::copy(V, model.set_real_variable("Dot_ux1"));
  gmm::resize(V, m1f_ui1.nb_dof()); gmm::clear(V); gmm::copy(V, model.set_real_variable("Dot_uz1"));
  gmm::resize(V, m1f_ui2.nb_dof()); gmm::clear(V); gmm::copy(V, model.set_real_variable("Dot_ux2"));
  gmm::resize(V, m1f_ui2.nb_dof()); gmm::clear(V); gmm::copy(V, model.set_real_variable("Dot_uz2"));
  gmm::resize(V, m1f_uo1.nb_dof()); gmm::clear(V); gmm::copy(V, model.set_real_variable("Dot_uo1"));
  gmm::resize(V, m1f_uo2.nb_dof()); gmm::clear(V); gmm::copy(V, model.set_real_variable("Dot_uo2"));

  gmm::resize(V, m2f_ui.nb_dof()); gmm::clear(V); gmm::copy(V, model.set_real_variable("Dot_ui"));
  gmm::resize(V, m2f_uo.nb_dof()); gmm::clear(V); gmm::copy(V, model.set_real_variable("Dot_uo"));

  plain_vector A;

  gmm::resize(A, m1f_ui1.nb_dof()); gmm::clear(A); gmm::copy(A, model.set_real_variable("Dot2_ux1"));
  gmm::resize(A, m1f_ui1.nb_dof()); gmm::clear(A); gmm::copy(A, model.set_real_variable("Dot2_uz1"));
  gmm::resize(A, m1f_ui2.nb_dof()); gmm::clear(A); gmm::copy(A, model.set_real_variable("Dot2_ux2"));
  gmm::resize(A, m1f_ui2.nb_dof()); gmm::clear(A); gmm::copy(A, model.set_real_variable("Dot2_uz2"));
  gmm::resize(A, m1f_uo1.nb_dof()); gmm::clear(A); gmm::copy(A, model.set_real_variable("Dot2_uo1"));
  gmm::resize(A, m1f_uo2.nb_dof()); gmm::clear(A); gmm::copy(A, model.set_real_variable("Dot2_uo2"));

  gmm::resize(A, m2f_ui.nb_dof()); gmm::clear(A); gmm::copy(A, model.set_real_variable("Dot2_ui"));
  gmm::resize(A, m2f_uo.nb_dof()); gmm::clear(A); gmm::copy(A, model.set_real_variable("Dot2_uo"));

  gmm::iteration iter(residual, 0, 40000);
  iter.set_noisy(2);

  // freeze variable
  model.disable_variable("ux1");
  model.disable_variable("Dot_ux1");
  model.disable_variable("Dot2_ux1");
  model.disable_variable("uz1");
  model.disable_variable("Dot_uz1");
  model.disable_variable("Dot2_uz1");
  model.disable_variable("uo1");
  model.disable_variable("Dot_uo1");
  model.disable_variable("Dot2_uo1");
  model.disable_variable("ux2");
  model.disable_variable("Dot_ux2");
  model.disable_variable("Dot2_ux2");
  model.disable_variable("uz2");
  model.disable_variable("Dot_uz2");
  model.disable_variable("Dot2_uz2");
  model.disable_variable("uo2");
  model.disable_variable("Dot_uo2");
  model.disable_variable("Dot2_uo2");
  model.disable_variable("ui");
  model.disable_variable("Dot_ui");
  model.disable_variable("Dot2_ui");
  model.disable_variable("uo");
  model.disable_variable("Dot_uo");
  model.disable_variable("Dot2_uo");

  // TOP boundary source term.
  std::vector<scalar_type> F2i(m2f_rhsi.nb_dof()); gmm::clear(F2i);
  std::vector<scalar_type> F2o(m2f_rhso.nb_dof()); gmm::clear(F2o);
  std::vector<scalar_type> F1i(m1f_rhsi.nb_dof()); gmm::clear(F1i);
  std::vector<scalar_type> F1o(m1f_rhso.nb_dof()); gmm::clear(F1o);
  model.add_initialized_fem_data("top_force_ui", m2f_rhsi, F2i);
  model.add_initialized_fem_data("top_force_uo", m2f_rhso, F2o);
  model.add_initialized_fem_data("top_force_ux1", m1f_rhsi, F1i);
  model.add_initialized_fem_data("top_force_uz1", m1f_rhsi, F1i);
  getfem::add_source_term_brick(model, mim, "ui", "top_force_ui", TOP);
  getfem::add_source_term_brick(model, mim, "uo", "top_force_uo", TOP);

  sol_t = 0.;

  // Iterations in time
  cout << "Iterations in time start\n";
  for (scalar_type t = 0.; t < T; t += dt) {

    sol_t = t+dt;

    cout << "solving for t = " << sol_t << endl;
    // Add source term on TOP boundary

    // Left
    model.enable_variable("ux1");
    model.enable_variable("Dot_ux1");
    model.enable_variable("Dot2_ux1");
    model.enable_variable("uz1");
    model.enable_variable("Dot_uz1");
    model.enable_variable("Dot2_uz1");
    model.enable_variable("uo1");
    model.enable_variable("Dot_uo1");
    model.enable_variable("Dot2_uo1");

    // Build all for debugging
    model.assembly(getfem::model::build_version(3));
    gmm::csc_matrix<double> MM;
    gmm::copy(model.real_tangent_matrix(), MM);
    gmm::MatrixMarket_save("MM.mtx", MM);
    gmm::vecsave("FF.vec", model.real_rhs());

    iter.init();
    //getfem::standard_solve(model, iter);

    model.disable_variable("ux1");
    model.disable_variable("Dot_ux1");
    model.disable_variable("Dot2_ux1");
    model.disable_variable("uz1");
    model.disable_variable("Dot_uz1");
    model.disable_variable("Dot2_uz1");
    model.disable_variable("uo1");
    model.disable_variable("Dot_uo1");
    model.disable_variable("Dot2_uo1");

    // Right
    model.enable_variable("ux2");
    model.enable_variable("Dot_ux2");
    model.enable_variable("Dot2_ux2");
    model.enable_variable("uz2");
    model.enable_variable("Dot_uz2");
    model.enable_variable("Dot2_uz2");
    model.enable_variable("uo2");
    model.enable_variable("Dot_uo2");
    model.enable_variable("Dot2_uo2");

    iter.init();
    //getfem::standard_solve(model, iter);

    model.disable_variable("ux2");
    model.disable_variable("Dot_ux2");
    model.disable_variable("Dot2_ux2");
    model.disable_variable("uz2");
    model.disable_variable("Dot_uz2");
    model.disable_variable("Dot2_uz2");
    model.disable_variable("uo2");
    model.disable_variable("Dot_uo2");
    model.disable_variable("Dot2_uo2");

    // Side Boundary

    model.enable_variable("ui");
    model.enable_variable("Dot_ui");
    model.enable_variable("Dot2_ui");
    model.enable_variable("uo");
    model.enable_variable("Dot_uo");
    model.enable_variable("Dot2_uo");

    // TODO: Add Boundary for of side

    iter.init();
    //getfem::standard_solve(model, iter);

    model.disable_variable("ui");
    model.disable_variable("Dot_ui");
    model.disable_variable("Dot2_ui");
    model.disable_variable("uo");
    model.disable_variable("Dot_uo");
    model.disable_variable("Dot2_uo");

    model.shift_variables_for_time_integration();

  }

  return (iter.converged());

}

/**************************************************************************/
/*  main program.                                                         */
/**************************************************************************/

int main(int argc, char *argv[]) {

  GMM_SET_EXCEPTION_DEBUG; // Exceptions make a memory fault, to debug.
  FE_ENABLE_EXCEPT;        // Enable floating point exception for Nan.

  try {
    elastodynamic_side_boundary_problem p;
    p.PARAM.read_command_line(argc, argv);
    p.init();
    p.solve();
  }
  GMM_STANDARD_CATCH_ERROR;

  return 0;
}
